/*******************************************************
Title:  MBUS driver library for Nokia Phones
Author: Roland Praml <praml.roland@t-online.de> www.praml.net
Date:   September 2002
Hardware: AVR AT90S8515 (with UART)

refer to the gnokii project or http://www.flosys.com/gsm/
for protocol details.

Nokia Handy connected to RX/TX pins:

RX-------+--> MBUS
TX--|<|--+

GND--------->
********************************************************/

#include <io.h>
#include "mbus.h"
#include <interrupt.h>
#include <sig-avr.h>
#define TRANSMIT outp((1<<TXEN)|(1<<CHR9),UCR);
#define RECEIVE  outp((1<<RXCIE)|(1<<RXEN)|(1<<CHR9),UCR);

static volatile u08 *uart_data_ptr;
static volatile int uart_counter;

u08 CRC;

u08 state; // Zustand

static volatile u08 source; // Quelle
static volatile u08 dest; // Ziel
static volatile u08 command; // Befehl

u08 buffer_ptr; // Hilfsvariable
static volatile u08 seq_id; // sequenz - ID
static volatile u08 frame_avaiable; 
u08 seqnum; // internal sequence counter


SIGNAL(SIG_UART_DATA)
/* signal handler for uart txd ready interrupt */
/* send next bit */
{
  unsigned char data;
  unsigned char parity;
  if (--uart_counter > 0) {
    data = *(uart_data_ptr++);

    parity = data  ^ (data >> 4);
    parity ^= parity >> 2;
    parity ^= parity >> 1;
    parity &= 1;
    CRC ^= data;

    if (parity) { /* set parity */
      cbi(UCR,TXB8);
    } else {
      sbi(UCR,TXB8);
    }
    outp(data,UDR); /* write byte to data buffer */
  } else {
    cbi(UCR,UDRIE);
    uart_counter = 0;
  }
}


void uart_send(u08 *buf, u08 size)
/* send buffer to uart */
{
  /* write first byte to data buffer */
  uart_data_ptr = buf;
  uart_counter = size + 1;
  sbi(UCR,UDRIE);
  while (uart_counter);
}


SIGNAL(SIG_UART_RECV) {
  u08 FrameHDR[6];
  u16 i;
  
  u08 usr  = inp(USR);
  u08 data = inp(UDR);
  u08 parity = data  ^(data >> 4);
  parity ^= parity >> 2;
  parity ^= parity >> 1;
  parity &= 1;
  if (bit_is_clear(UCR,RXB8)) parity ^= 1;
  
  if (usr & (BV(FE)|BV(OVR)) ) {
    state = 0; // reset the automat
  } else if (parity) {
    state = 0;
  } else {
    switch (state) {
      
    case 0: // frame start
      if (data ==  0x1F) {
	state++;
	CRC = 0;
      }
      break;
      
    case 1: // dest
      dest = data;
      if (data == 0x00) {
	state++;
      } else if (data == 0x1D) {
	state++;
      } else state = 0;
      break;
      
    case 2: // source
      source = data;
      if (data == 0x00) {
	state++;
      } else if (data == 0x1D) {
	state++;
      } else state = 0;
      break;
      
    case 3: // command
      command = data;
      state++;
      if (data == 0x7F) {
	state = 7; // ACK
      }
      break;
      
    case 4: // always 0x00
      state++;
      if (data != 0x00) state = 0;
      break;
      
    case 5: // length
      buffer_ptr = 0;
      MBUS_bufsize = data;
      state++;
      if (data == 0) state++; // no Frame
      break;
      
    case 6: // read Data, max 16 Byte
      if (buffer_ptr < 15) {
	MBUS_buffer[buffer_ptr] = data;
      }
      if (buffer_ptr < 100) buffer_ptr++;
      if (buffer_ptr >= MBUS_bufsize) state++;
      break;
      
    case 7: // seq_id
      seq_id = data;
      state++;
      break;
      
    case 8: //CRC
      //cbi(UCR,RXCIE); // Empfang abschalten;
      if (CRC == data) {
	if (command != 0x7f) {
	  // send ACK
	  for (i = 0; i < 3000; i++) asm("nop");
	  FrameHDR[0] = 0x1F; //frame Start
	  FrameHDR[1] = 0x00; // dest =  Mobile Phone
	  FrameHDR[2] = 0x1D; // source = PC
	  FrameHDR[3] = 0x7F; // ack
	  FrameHDR[4] = seq_id;
	  FrameHDR[5] = 0x7D  ^seq_id; // CRC
	  sei();
	  TRANSMIT;
	  uart_send(FrameHDR, 6); // send ack
	  RECEIVE;
	}
	frame_avaiable = 1;
      }
      state = 0;
      break;
    }
    CRC ^= data;
  }
}


/* receive 0ne frame, return 0 if timeout */
u08 uart_receive(void) {
  u16 timeout = 0;
  u08 delay;

  while (!frame_avaiable) {
    for (delay = 0; delay < 255; delay++) asm("nop;");
    timeout++;
    if (timeout > 20000) {
      frame_avaiable = 0;  
      return 0;
    }
  }
  frame_avaiable = 0;  
  return 1;
}

/* waits for ack of Mobile */
u08 wait_for_ack(void) {
  if (!uart_receive()) return 0;
  if (source != 0x00) return 0; // mobile
  if (dest != 0x1D) return 0; // PC 
  if (command != 0x7f) return 0; // Ack-cmd
  if (seq_id != seqnum) return 0; // not the ack for sent frame
  return 1;
}

/* send a frame Command cmd
   data & length of data and wait for ack, returns 1 if successful */
u08 MBUS_send_frame(u08 cmd, u08 *buf, u08 size) {
  //u16 i;
  u08 FrameHDR[6];
  frame_avaiable = 0;
  //for (i = 0; i < 6000; i++) asm("nop");  
  FrameHDR[0] = 0x1F; //frame Start
  FrameHDR[1] = 0x00; // dest =  Mobile Phone
  FrameHDR[2] = 0x1D; // source = PC
  FrameHDR[3] = cmd;
  FrameHDR[4] = 0x00;
  FrameHDR[5] = size;
  CRC = 0; // reset CRC
  TRANSMIT; // enable Transmitter
  uart_send(FrameHDR, 6);
  uart_send(buf, size);
  seqnum++;
  uart_send(&seqnum, 1);
  uart_send(&CRC, 1);
  CRC = 0;
  state = 0;
  RECEIVE; // switch to receive mode
  return wait_for_ack(); // wait for ack packet.
}

u08 MBUS_recv_frame(u08 cmd) {
  CRC = 0;
  state = 0;
  RECEIVE;
  if (!uart_receive()) return 0;
  if (source != 0x00) return 0; // mobile
  if (dest != 0x1D) return 0; // PC
  if (command != cmd) return 0; // is this the frame we want?
  return 1;
}
/* send and receive one frame */
u08 MBUS_command (u08 cmd, u08 *buf, u08 size) {
  if (MBUS_send_frame(cmd, buf, size)) {
	return MBUS_recv_frame(cmd);
      }
      return 0; // no success
}

void MBUS_init(void) {
    /* set baud rate to 9600 */
    outp(25,UBRR);
    sbi(PORTD,PD0); /* pullup am Eingang */
}


/*
MBUS_init: Init Ports and Baud_rate

MBUS_command(cmd, *bus, size): Send Command cmd with payload buf & size
*/

